unit KASteganographyImage;

interface

uses
  Windows, SysUtils, Classes, Graphics, Controls, ExtCtrls;

type
  TStActionType=(atDecode,atEncode);
  TOnProgressEvent            = Procedure (Sender : TObject; Current, Total : Integer) of Object;
  TOnImageProcessingDoneEvent = Procedure (Sender : TObject; MustSave : Boolean; Var Abort : Boolean) of Object;
  TOnNextImageEvent           = Procedure (Sender : TObject; ImageNumber : Integer; Var Abort : Boolean) of Object;


  TKASteganographyImage = class(TImage)
  private
    FPassword                 : String;
    FEmbeddedText             : TStrings;
    FActivate                 : Boolean;
    FActionType               : TStActionType;

    FEmbeddedStream           : TMemoryStream;
    FMarkerPos                : Integer;
    FSizePos                  : Integer;
    FCRCPos                   : Integer;
    FPasswordPos              : Integer;
    FImagePos                 : Integer;
    FImageNumber              : Integer;
    FCRCTable                 : array[0..512] Of Longint;
    FOnProgress               : TOnProgressEvent;
    FOnImageProcessingDone    : TOnImageProcessingDoneEvent;
    FOnNeedNextImage          : TOnNextImageEvent;
    FVersion                  : String;
    FMultiEmbedding           : Boolean;

    procedure SetEmbeddedText(const Value: TStrings);
    procedure SetPassword(const Value: String);
    procedure SetActivate(const Value: Boolean);
    procedure SetActionType(const Value: TStActionType);
    procedure SetOnProgress(const Value: TOnProgressEvent);
    procedure SetDecoded(const Value: Boolean);
    function  GetDecoded: Boolean;
    procedure SetVersion(const Value: String);
    procedure SetOnNeedNextImage(const Value: TOnNextImageEvent);
    procedure SetOnImageProcessingDone(const Value: TOnImageProcessingDoneEvent);
    function GetImageSequenceNo: Integer;
    procedure SetMultiEmbedding(const Value: Boolean);
    { Private declarations }
  protected
    { Protected declarations }
    procedure BuildCRCTable;
    function  RecalcCRC(B : Byte; CrcOld: Longint) : Cardinal;
    function  CalcPasswordCRC : Cardinal;
    function  GetNextByte(var ByteData : Byte) : Boolean;
    function  GetHeaderPart(var X, Y : Integer; Var R, G, B, CurrentByte : Byte):Cardinal;
    Procedure EmbedInImage;
    Procedure ParseImage;
  public
    { Public declarations }
    Procedure   SetEmbeddedData(Data : TStream);Overload;
    Procedure   SetEmbeddedData(Data : TStrings);Overload;
    Procedure   SetEmbeddedData(Data : String);Overload;
    Function    GetAsStream          : TStream;
    Function    GetAsStringList      : TStringList;
    Function    GetAsString          : String;
    Function    GetMaxEmbedSzie      : Integer;

    Function    HasEmbeddedData      : Boolean;
    Function    EmbeddedDataSize     : Integer;
    function    CalcCRC(AString : String) : Cardinal;
    Function    ValidPassword(Password:String):Boolean;
    Procedure   Clear;
    Procedure   LoadFromFile(const Filename: string);
    Procedure   SaveToFile(const Filename: string);
    Property    ImageSequenceNo : Integer read GetImageSequenceNo;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy;override;
  published
    { Published declarations }
    Property Version                 : String                      read FVersion               write SetVersion;
    Property Password                : String                      read FPassword              write SetPassword;
    Property EmbeddedText            : TStrings                    read FEmbeddedText          write SetEmbeddedText;
    Property Decoded                 : Boolean                     read GetDecoded             write SetDecoded         Default False;
    Property MultiEmbedding          : Boolean                     read FMultiEmbedding        write SetMultiEmbedding  Default True;
    Property ActionType              : TStActionType               read FActionType            write SetActionType      Default atDecode;
    Property Activate                : Boolean                     read FActivate              write SetActivate        Default False;
    Property OnProgress              : TOnProgressEvent            read FOnProgress            write SetOnProgress;
    Property OnImageProcessingDone   : TOnImageProcessingDoneEvent read FOnImageProcessingDone write SetOnImageProcessingDone;
    Property OnNeedNextImage         : TOnNextImageEvent           read FOnNeedNextImage       write SetOnNeedNextImage;
  end;

procedure Register;

implementation
Const
  CRCPOLY    = $EDB88320;
  KAMARKER   = 64052467;
  HEADERSIZE = SizeOf(Integer)*4;

procedure Register;
begin
  RegisterComponents('KA', [TKASteganographyImage]);
end;

{ TKASteganographyImage }

constructor TKASteganographyImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPassword                  := '';
  FVersion                   := '3.0';
  FMarkerPos                 := 1;
  FSizePos                   := 1;
  FCRCPos                    := 1;
  FPasswordPos               := 1;
  FImagePos                  := 1;
  FImageNumber               := 1;
  FMultiEmbedding            := True;
  FOnProgress                := Nil;
  FOnImageProcessingDone     := Nil;
  FOnNeedNextImage           := Nil;
  FEmbeddedText              := TStringList.Create;
  FEmbeddedStream            := TMemoryStream.Create;
  BuildCRCTable;
end;

destructor TKASteganographyImage.Destroy;
begin
  FEmbeddedStream.Free;
  FEmbeddedText.Free;
  inherited Destroy;
end;

procedure TKASteganographyImage.BuildCRCTable;
var
  I : Word;
  J : Word;
  R : Cardinal;
begin
  FillChar(FCRCTable, SizeOf(FCRCTable), 0);
  For I := 0 To 255 do
      Begin
        R := I SHL 1;
        For J := 8 Downto 0 Do
          Begin
            if (R AND 1) <> 0 Then
               R := (R SHR 1) XOR CRCPOLY
            Else
               R := R SHR 1;
          End;
        FCRCTable[I] := R;
      End;
end;

function TKASteganographyImage.RecalcCRC(B : Byte; CrcOld: Longint): Cardinal;
begin
  Result := FCRCTable[Byte(CrcOld XOR Longint(B))] XOR ((CrcOld SHR 8) and $00FFFFFF)
end;

function TKASteganographyImage.CalcCRC(AString : String) : Cardinal;
Var
  X : Integer;
Begin
  Result := $FFFFFFFF;
  if Length(AString) > 0 Then
     Begin
        For X := 1 To Length(AString) do
            Begin
              Result := RecalcCRC(ORD(AString[X]),Result);
            End;
     End;
End;

function TKASteganographyImage.CalcPasswordCRC : Cardinal;
Begin
  Result := CalcCRC(FPassword);
End;

function TKASteganographyImage.GetNextByte(var ByteData : Byte) : Boolean;
Var
  SZ  : Integer;
  SZA : Array [1..4] of Byte Absolute SZ;
begin
  Result := False;
  if FMarkerPos < 5 Then
     Begin
       SZ       := KAMARKER;
       ByteData := SZA[FMarkerPos];
       Inc(FMarkerPos);
       Result   := True;
     End
  Else
  if FCRCPos < 5 Then
     Begin
       SZ       := CalcPasswordCRC;
       ByteData := SZA[FCRCPos];
       Inc(FCRCPos);
       Result   := True;
     End
  Else
  if FSizePos < 5 Then
     Begin
       SZ       := FEmbeddedStream.Size;
       ByteData := SZA[FSizePos];
       Inc(FSizePos);
       Result   := True;
     End
  Else
  if FImagePos < 5 Then
     Begin
       SZ       := FImageNumber;
       ByteData := SZA[FImagePos];
       Inc(FImagePos);
       Result   := True;
     End
  Else
  if FEmbeddedStream.Read(ByteData,1)=1 Then
     Begin
       ByteData := ByteData XOR Byte(FPassword[FPasswordPos]);
       Inc(FPasswordPos);
       if FPasswordPos > Length(FPassword) Then FPasswordPos := 1;
       Result := True;
     End;
end;

procedure TKASteganographyImage.EmbedInImage;
Var
  MaxEmbeddedData : Integer;
  DataByte        : Byte;
  X               : Integer;
  Y               : Integer;
  J               : Integer;
  R               : Byte;
  G               : Byte;
  B               : Byte;
  ByteMask        : Byte;
  BitData         : Byte;
  PixelData       : Byte;
  CurrentByte     : Byte;
  Abort           : Boolean;
  MustSave        : Boolean;
begin
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty)
  And (Password<>'')
  And (FEmbeddedStream.Size > 0) Then
      Begin
        MaxEmbeddedData     := (Picture.Bitmap.Width*Picture.Bitmap.Height*3) DIV 8;
        if MaxEmbeddedData  < HEADERSIZE Then
           Begin
             Raise Exception.Create('Image is too small for holding any data!');
           End;
        if MaxEmbeddedData < FEmbeddedStream.Size+HEADERSIZE Then
           Begin
             MaxEmbeddedData := MaxEmbeddedData-HEADERSIZE;
             if (NOT Assigned(FOnNeedNextImage)) OR (NOT FMultiEmbedding) Then
                Begin
                  Raise Exception.Create('Data to embed is too much for the image! Maximum embedded data size is '+IntToStr(MaxEmbeddedData)+' bytes.');
                End;
           End;
        X                           := 0;
        Y                           := 0;
        FMarkerPos                  := 1;
        FCRCPos                     := 1;
        FSizePos                    := 1;
        FPasswordPos                := 1;
        FImagePos                   := 1;
        FImageNumber                := 1;
        Picture.Bitmap.PixelFormat  := pf24bit;
        Abort                       := False;
        FEmbeddedStream.Position    := 0;
        CurrentByte                 := 0;
        R                           := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        G                           := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        B                           := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        While GetNextByte(DataByte) Do
             Begin
               ByteMask    := $80;
               For J := 1 To 8 do
                   Begin
                     BitData := DataByte and ByteMask;
                     if CurrentByte=0 Then
                        Begin
                          R := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                          G := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                          B := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                        End;
                     if CurrentByte=0 Then
                        PixelData := R
                     Else
                     if CurrentByte=1 Then
                        PixelData := G
                     Else
                        PixelData := B;

                     if BitData > 0 Then
                        Begin
                          if (PixelData And 1) = 0 Then
                             Begin
                               Inc(PixelData);
                             End;
                        End
                     Else
                        Begin
                          if (PixelData And 1) > 0 Then
                             Begin
                               Dec(PixelData);
                             End;
                        End;

                     if CurrentByte=0 Then
                        R := PixelData
                     Else
                     if CurrentByte=1 Then
                        G  := PixelData
                     Else
                     if CurrentByte=2 Then
                        B  := PixelData;

                     Inc(CurrentByte);
                     if (CurrentByte=3) Then
                        Begin
                          Picture.Bitmap.Canvas.Pixels[X, Y] := TColor(RGB(R,G,B));
                          X := (X + 1) Mod Picture.Bitmap.Width;
                          if (X = 0) Then Inc(Y);
                          CurrentByte := 0;
                        End;

                     ByteMask := ByteMask SHR 1;
                     if Assigned(FOnProgress) Then
                        Begin
                          FOnProgress(Self,FEmbeddedStream.Position,FEmbeddedStream.Size);
                        End;

                     if (Y = Picture.Bitmap.Height) And (FEmbeddedStream.Position < FEmbeddedStream.Size) Then
                        Begin
                          MustSave := True;
                          if Assigned(FOnImageProcessingDone) Then
                             Begin
                               FOnImageProcessingDone(Self,MustSave,Abort);
                             End;
                          if Abort Then Exit;

                          if (Assigned(FOnNeedNextImage)) And (FMultiEmbedding) Then
                             Begin
                               FMarkerPos       := 1;
                               FCRCPos          := 1;
                               FSizePos         := 1;
                               FImagePos        := 1;
                               X                := 0;
                               Y                := 0;
                               CurrentByte      := 0;
                               Inc(FImageNumber);
                               Repeat
                                 FOnNeedNextImage(Self,FImageNumber,Abort);
                                 MaxEmbeddedData     := (Picture.Bitmap.Width*Picture.Bitmap.Height*3) DIV 8;
                               Until (NOT Picture.Bitmap.Empty) And (MaxEmbeddedData > HEADERSIZE);
                             End
                          Else
                             Begin
                               Raise Exception.Create('Data to embed is too much for the image. Maximum embedded data size is '+IntToStr(MaxEmbeddedData)+' bytes.');
                             End;
                          if Abort Then Exit;
                        End;
                   End;
             End;
        if (CurrentByte > 0) Then
           Begin
             Picture.Bitmap.Canvas.Pixels[X, Y] := TColor(RGB(R,G,B));
           End;
        if Assigned(FOnImageProcessingDone) Then
           Begin
             MustSave := FImageNumber > 1;
             FOnImageProcessingDone(Self, MustSave, Abort);
           End;
      End
  Else
      Begin
        if  NOT (Assigned(Picture))         Then  Raise Exception.Create('No bitmap loaded!');
        if  NOT (Assigned(Picture.Bitmap))  Then  Raise Exception.Create('No bitmap loaded!');
        if      (Picture.Bitmap.Empty)      Then  Raise Exception.Create('No bitmap loaded or image is not a windows bitmap!');
        if  NOT (Password<>'')              Then  Raise Exception.Create('Password property must be filled!');
        if  NOT (FEmbeddedStream.Size > 0)  Then  Raise Exception.Create('There is no data to encode!');
      End;
end;

function TKASteganographyImage.GetMaxEmbedSzie: Integer;
begin
  Result := 0;
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty) Then
      Begin
        Result := ((Picture.Bitmap.Width*Picture.Bitmap.Height*3) DIV 8)-HEADERSIZE;
      End;
end;

function TKASteganographyImage.GetHeaderPart(var X, Y : Integer; Var R, G, B, CurrentByte : Byte): Cardinal;
Var
  I          : Integer;
  J          : Integer;
  ByteData   : Byte;
  PixelData  : Byte;
  SZ         : Integer;
  SZA        : Array [1..4] of Byte Absolute SZ;
begin
  SZ     := 0;
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty) Then
      Begin
        For I := 1 to 4 do
            Begin
              ByteData := 0;
              For J := 8 Downto 1 do
                  Begin
                    if CurrentByte=0 Then
                       Begin
                          R := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                          G := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                          B := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                       End;
                    if CurrentByte=0 Then
                        PixelData := R
                    Else
                    if CurrentByte=1 Then
                       PixelData := G
                    Else
                       PixelData := B;
                    ByteData := ByteData + (Byte(PixelData AND $00000001) SHL (J-1));
                    Inc(CurrentByte);
                    if (CurrentByte=3) Then
                       Begin
                         X := (X + 1) Mod Picture.Bitmap.Width;
                         if   (X = 0) Then Inc(Y);
                         CurrentByte := 0;
                       End;
                  End;
              SZA[I] := ByteData;
            End;
      End;
  Result := Sz;
end;

function TKASteganographyImage.HasEmbeddedData: Boolean;
Var
  X               : Integer;
  Y               : Integer;
  R               : Byte;
  G               : Byte;
  B               : Byte;
  CurrentByte     : Byte;
  SZ              : Integer;
  SZA             : Array [1..4] of Byte Absolute SZ;
  FileMarker      : Integer;
begin
  Result := False;
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty) Then
      Begin
        X                          := 0;
        Y                          := 0;
        CurrentByte                := 0;
        Picture.Bitmap.PixelFormat := pf24Bit;
        R                          := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        G                          := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        B                          := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        //*******************************************************  MARKER_HEADER
        FileMarker := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        if FileMarker = KAMARKER Then Result := True;
      End;
end;

function TKASteganographyImage.ValidPassword(Password: String): Boolean;
Var
  X               : Integer;
  Y               : Integer;
  R               : Byte;
  G               : Byte;
  B               : Byte;
  CurrentByte     : Byte;
  SZ              : Integer;
  SZA             : Array [1..4] of Byte Absolute SZ;
  FileMarker      : Integer;
  CRC32           : Cardinal;
begin
  Result := False;
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty) Then
      Begin
        X                          := 0;
        Y                          := 0;
        Picture.Bitmap.PixelFormat := pf24Bit;
        CurrentByte                := 0;
        R                          := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        G                          := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        B                          := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        //*******************************************************  MARKER_HEADER
        FileMarker := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        if FileMarker <> KAMARKER Then Exit;
        //******************************************************  PASSWORD_CRC32
        CRC32      := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        if CalcCRC(Password) = CRC32 Then Result := True;
      End;
end;


procedure TKASteganographyImage.ParseImage;
Var
  X               : Integer;
  Y               : Integer;
  I               : Integer;
  J               : Integer;
  R               : Byte;
  G               : Byte;
  B               : Byte;
  ByteData        : Byte;
  PixelData       : Byte;
  CurrentByte     : Byte;
  SZ              : Integer;
  SZA             : Array [1..4] of Byte Absolute SZ;
  StreamSize      : Integer;
  TmpStreamSize   : Integer;
  TmpImageNumber  : Integer;
  CRC32           : Cardinal;
  FileMarker      : Integer;
  Abort           : Boolean;
  MustSave        : Boolean;
  NextLoadOK      : Boolean;
begin
  FEmbeddedStream.Position   := 0;
  FEmbeddedStream.Size       := 0;
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty)
  And (Password<>'') Then
      Begin
        X                          := 0;
        Y                          := 0;
        Picture.Bitmap.PixelFormat := pf24Bit;
        CurrentByte                := 0;
        R                          := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        G                          := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        B                          := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        //*******************************************************  MARKER_HEADER
        FileMarker := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        if FileMarker <> KAMARKER Then
           Begin
             Raise Exception.Create('There is no data embedded in the picture!');
           End;
        //******************************************************  PASSWORD_CRC32
        CRC32 := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        if CalcPasswordCRC <> CRC32 Then
           Begin
             Raise Exception.Create('Invalid password!');
           End;
        //****************************************************  STREAM_DATA_SIZE
        StreamSize := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        if StreamSize = 0 Then
           Begin
             Raise Exception.Create('There is no data embedded in the picture!');
           End;
        FEmbeddedStream.Size := StreamSize;
        //********************************************************  IMAGE_NUMBER
        FImageNumber := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        //*****************************************************  STREAM_CONTENTS
        FPasswordPos := 1;
        For I := 0 to StreamSize-1 do
            Begin
              ByteData    := 0;
              For J := 8 Downto 1 do
                  Begin
                    if CurrentByte=0 Then
                       Begin
                          R := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                          G := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                          B := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                       End;
                    if CurrentByte=0 Then
                        PixelData := R
                    Else
                    if CurrentByte=1 Then
                       PixelData := G
                    Else
                       PixelData := B;
                    ByteData := ByteData + (Byte(PixelData AND $00000001) SHL (J-1));
                    Inc(CurrentByte);
                    if (CurrentByte=3) Then
                       Begin
                         X := (X + 1) Mod Picture.Bitmap.Width;
                         if   (X = 0) Then Inc(Y);
                         CurrentByte := 0;
                       End;

                    if (Y = Picture.Bitmap.Height) And (I < StreamSize-1) Then
                       Begin
                         if Assigned(FOnImageProcessingDone) Then
                            Begin
                              MustSave := False;
                              FOnImageProcessingDone(Self,MustSave,Abort);
                            End;
                         if Abort Then Exit;
                         if Assigned(FOnNeedNextImage) Then
                            Begin
                              FImagePos := 1;
                              Inc(FImageNumber);
                              Repeat
                                NextLoadOK := False;
                                FOnNeedNextImage(Self,FImageNumber,Abort);
                                if (NOT Abort) And (NOT Picture.Bitmap.Empty) Then
                                   Begin
                                     X                          := 0;
                                     Y                          := 0;
                                     Picture.Bitmap.PixelFormat := pf24Bit;
                                     CurrentByte                := 0;
                                     R                          := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                                     G                          := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                                     B                          := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
                                     FileMarker                 := GetHeaderPart(X, Y, R, G, B, CurrentByte);
                                     CRC32                      := GetHeaderPart(X, Y, R, G, B, CurrentByte);
                                     TmpStreamSize              := GetHeaderPart(X, Y, R, G, B, CurrentByte);
                                     TmpImageNumber             := GetHeaderPart(X, Y, R, G, B, CurrentByte);
                                     NextLoadOK                 :=     (FileMarker      = KAMARKER)
                                                                   AND (CalcPasswordCRC = CRC32)
                                                                   AND (TmpStreamSize   = StreamSize)
                                                                   AND (TmpImageNumber  = FImageNumber);
                                   End;
                              Until (Abort) Or (NextLoadOK);
                            End
                         Else
                            Begin
                              Raise Exception.Create('Embedded data is too much for the image. You must handle OnNeedNextImage event!');
                            End;
                         if Abort Then Exit;
                      End;
                  End;
              ByteData := ByteData XOR Byte(FPassword[FPasswordPos]);
              Inc(FPasswordPos);
              if FPasswordPos > Length(FPassword) Then FPasswordPos := 1;
              FEmbeddedStream.Write(ByteData,1);
              if Assigned(FOnProgress) Then
                 Begin
                   FOnProgress(Self,I+1,StreamSize);
                 End;
            End;
        FEmbeddedStream.Position := 0;
      End
  Else
      Begin
        if  NOT (Assigned(Picture))         Then  Raise Exception.Create('No bitmap loaded!');
        if  NOT (Assigned(Picture.Bitmap))  Then  Raise Exception.Create('No bitmap loaded!');
        if      (Picture.Bitmap.Empty)      Then  Raise Exception.Create('No bitmap loaded or image is not a windows bitmap!');
        if  NOT (Password<>'')              Then  Raise Exception.Create('Password property must be filled!');
      End;
End;

function TKASteganographyImage.GetImageSequenceNo: Integer;
Var
  FileMarker  : Integer;
  CRC32       : Integer;
  StreamSize  : Integer;
  X           : Integer;
  Y           : Integer;
  R           : Byte;
  G           : Byte;
  B           : Byte;
  CurrentByte : Byte;
begin
  Result := 0;
  if  (Assigned(Picture))
  And (Assigned(Picture.Bitmap))
  And (NOT Picture.Bitmap.Empty) Then
     Begin
        X                          := 0;
        Y                          := 0;
        CurrentByte                := 0;
        R                          := GetRValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        G                          := GetGValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        B                          := GetBValue(Picture.Bitmap.Canvas.Pixels[X, Y]);
        FileMarker := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        CRC32      := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        StreamSize := GetHeaderPart(X, Y, R, G, B, CurrentByte);
        Result     := GetHeaderPart(X, Y, R, G, B, CurrentByte);
     End;
end;

procedure TKASteganographyImage.Clear;
begin
  EmbeddedText.Text        := '';
  FEmbeddedStream.Position := 0;
  FEmbeddedStream.Size     := 0;
  Activate                 := False;
  Self.Picture.Bitmap.Assign(Nil);
end;

function TKASteganographyImage.GetAsStream: TStream;
begin
  FEmbeddedStream.Position := 0;
  Result                   := FEmbeddedStream;
end;

function TKASteganographyImage.GetAsString: String;
begin
  Result                   := '';
  if FEmbeddedStream.Size > 0 Then
     Begin
       FEmbeddedStream.Position := 0;
       SetLength(Result,FEmbeddedStream.Size);
       FEmbeddedStream.Read(Result[1],FEmbeddedStream.Size);
       FEmbeddedStream.Position := 0;
     End;
end;

function TKASteganographyImage.GetAsStringList: TStringList;
begin
  FEmbeddedStream.Position := 0;
  FEmbeddedText.LoadFromStream(FEmbeddedStream);
  Result                   := TStringList(FEmbeddedText);
  FEmbeddedStream.Position := 0;
end;

procedure TKASteganographyImage.SetEmbeddedData(Data: TStream);
begin
  FEmbeddedStream.Position :=0;
  FEmbeddedStream.LoadFromStream(Data);
  FEmbeddedStream.Position := 0;
end;

procedure TKASteganographyImage.SetEmbeddedData(Data: TStrings);
begin
  FEmbeddedStream.Position :=0;
  Data.SaveToStream(FEmbeddedStream);
  FEmbeddedStream.Position :=0;
end;

procedure TKASteganographyImage.SetEmbeddedData(Data: String);
begin
  FEmbeddedStream.Position := 0;
  FEmbeddedStream.Write(Data[1],Length(Data));
  FEmbeddedStream.Position := 0;
end;

function TKASteganographyImage.EmbeddedDataSize: Integer;
begin
  Result := FEmbeddedStream.Size;
end;

procedure TKASteganographyImage.SetEmbeddedText(const Value: TStrings);
begin
  FEmbeddedText.Assign(Value);
  SetEmbeddedData(FEmbeddedText);
end;

procedure TKASteganographyImage.SetPassword(const Value: String);
begin
  FPassword := Trim(Value);
end;

procedure TKASteganographyImage.SetActionType(const Value: TStActionType);
begin
  FActionType := Value;
end;

procedure TKASteganographyImage.LoadFromFile(const Filename: string);
begin
  Picture.Bitmap.LoadFromFile(Filename);
  Picture.Bitmap.PixelFormat := pf24Bit;
end;

procedure TKASteganographyImage.SaveToFile(const Filename: string);
begin
  Picture.Bitmap.PixelFormat := pf24Bit;
  Picture.Bitmap.SaveToFile(Filename);
end;

procedure TKASteganographyImage.SetActivate(const Value: Boolean);
begin
  FActivate := Value;
  if FActivate Then
     Begin
       Try
         Case FActionType of
           atEncode : EmbedInImage;
           atDecode : ParseImage;
         End;
         if csDesigning in ComponentState Then MessageBox(GetFocus,'Success executing Steganography command!','Success',MB_ICONEXCLAMATION OR MB_OK);
       Finally
         if csDesigning in ComponentState Then FActivate := False;
       End;
     End
end;

procedure TKASteganographyImage.SetOnProgress(const Value: TOnProgressEvent);
begin
  FOnProgress := Value;
end;

procedure TKASteganographyImage.SetDecoded(const Value: Boolean);
begin
  //************************************************* READ ONLY
end;

procedure TKASteganographyImage.SetVersion(const Value: String);
begin
  //************************************************* READ ONLY
end;

function TKASteganographyImage.GetDecoded: Boolean;
begin
  Result := FEmbeddedStream.Size > 0;
end;

procedure TKASteganographyImage.SetOnNeedNextImage(
  const Value: TOnNextImageEvent);
begin
  FOnNeedNextImage := Value;
end;

procedure TKASteganographyImage.SetOnImageProcessingDone(
  const Value: TOnImageProcessingDoneEvent);
begin
  FOnImageProcessingDone := Value;
end;

procedure TKASteganographyImage.SetMultiEmbedding(const Value: Boolean);
begin
  FMultiEmbedding := Value;
end;



end.
